/* hinting-viewer -- View font outline hinting
 * Copyright 2005, Red Hat
 * Author: Owen Taylor <otaylor@redhat.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#define PANGO_ENABLE_BACKEND	/* Needed to access PangoFcFont.font_pattern */

#include <math.h>

#include <pango/pangoft2.h>
#include <cairo.h>

#include FT_OUTLINE_H

#include "render.h"

typedef enum
{
  RENDER_MODE_FILL,
  RENDER_MODE_STROKE,
  RENDER_MODE_DRAW_POINTS,
  RENDER_MODE_DRAW_CONTROL
} RenderMode;

struct _FontRenderer
{
  gdouble scale;
  gdouble point_radius;

  cairo_t *context;
  RenderMode mode;
};

static void
current_point_to_origin (cairo_t *context)
{
  double x, y;

  cairo_get_current_point (context, &x, &y);
  cairo_translate (context, x, y);
}

static guint32
get_load_flags (PangoFont *font)
{
  FcPattern *pattern = PANGO_FC_FONT (font)->font_pattern;
  FcBool antialias, hinting, autohint;
  guint32 load_flags = 0;

  /* disable antialiasing if requested */
  if (FcPatternGetBool (pattern,
			FC_ANTIALIAS, 0, &antialias) != FcResultMatch)
    antialias = FcTrue;
  
  if (!antialias)
    load_flags |= FT_LOAD_NO_BITMAP;
  else
    load_flags |= FT_LOAD_TARGET_MONO;
  
  /* disable hinting if requested */
  if (FcPatternGetBool (pattern,
			FC_HINTING, 0, &hinting) != FcResultMatch)
    hinting = FcTrue;
  
  if (!hinting)
    load_flags |= FT_LOAD_NO_HINTING;
  
  /* force autohinting if requested */
  if (FcPatternGetBool (pattern,
			FC_AUTOHINT, 0, &autohint) != FcResultMatch)
    autohint = FcFalse;
  
  if (autohint)
    load_flags |= FT_LOAD_FORCE_AUTOHINT;

  return load_flags;
}

static void
move_to_pango (FontRenderer *renderer,
	       cairo_t      *context,
	       int           x,
	       int           y)
{
  cairo_move_to (context,
		 (renderer->scale * x) / PANGO_SCALE,
		 (renderer->scale * y) / PANGO_SCALE);
}

static void
rel_move_to_pango (FontRenderer *renderer,
		   cairo_t      *context,
		   int           x,
		   int           y)
{
  cairo_move_to (context,
		 (renderer->scale * x) / PANGO_SCALE,
		 (renderer->scale * y) / PANGO_SCALE);
}

#define SCALE(pos) (renderer->scale * (pos) / 64.)
#define POINT(vector) SCALE((vector)->x), SCALE(-(vector)->y)
  
static int
draw_outline_move_to (FT_Vector *to,
		      void      *user)
{
  FontRenderer *renderer = user;

  cairo_move_to (renderer->context, POINT (to));

  return 0;
}

static int
draw_outline_line_to (FT_Vector *to,
		      void      *user)
{
  FontRenderer *renderer = user;

  cairo_line_to (renderer->context, POINT (to));

  return 0;
}

static int
draw_outline_conic_to (FT_Vector *control,
		       FT_Vector *to,
		       void      *user)
{
  FontRenderer *renderer = user;
  double last_x, last_y;
  double control_x, control_y;
  double to_x, to_y;

  cairo_get_current_point (renderer->context, &last_x, &last_y);

  control_x = SCALE (control->x);
  control_y = SCALE (- control->y);
  to_x = SCALE (to->x);
  to_y = SCALE (- to->y);

  cairo_curve_to (renderer->context,
		  last_x * (1./3.) + control_x * (2./3.),
		  last_y * (1./3.) + control_y * (2./3.),
		  to_x * (1./3.)   + control_x * (2./3.),
		  to_y * (1./3.)   + control_y * (2./3.),
		  to_x,
		  to_y);

  return 0;
}

static int
draw_outline_cubic_to (FT_Vector *control1,
		       FT_Vector *control2,
		       FT_Vector *to,
		       void      *user)
{
  FontRenderer *renderer = user;

  cairo_curve_to (renderer->context,
		  POINT(control1), POINT(control2), POINT (to));

  return 0;
}

static const FT_Outline_Funcs draw_outline_funcs = {
  draw_outline_move_to,
  draw_outline_line_to,
  draw_outline_conic_to,
  draw_outline_cubic_to,
  0,
  0,
};

static void
draw_outline (FontRenderer *renderer,
	      cairo_t      *context,
	      FT_Outline   *outline)
{
  cairo_save (context);
  current_point_to_origin (context);

  cairo_new_path (context);
  FT_Outline_Decompose (outline, &draw_outline_funcs, renderer);

  if (renderer->mode == RENDER_MODE_FILL)
    cairo_fill (context);
  else
    cairo_stroke (context);
  
  cairo_restore (context);
}

/* Draw a line from x1 to x2 terminated with a circle */
static void
draw_control (FontRenderer *renderer,
	      cairo_t      *context,
	      double        x1,
	      double        y1,
	      double        x2,
	      double        y2,
	      gboolean      draw_circle)
{
  double radius = renderer->point_radius;
  double dx = x2 - x1;
  double dy = y2 - y1;
  double nx2, ny2;
  double distance = sqrt (dx*dx + dy+dy);

  if (distance == 0)
    return;

  dx /= distance;
  dy /= distance;

  nx2 = x2 - dx * radius;
  ny2 = y2 - dy * radius;

  cairo_move_to (context, x1, y1);
  cairo_line_to (context, x2, y2);
  cairo_stroke (context);
  
  if (draw_circle)
    cairo_arc (context, x2, y2, radius, 0, 2 * M_PI);

  cairo_stroke (context);
}

typedef struct _DrawPointsState DrawPointsState;

struct _DrawPointsState
{
  FontRenderer *renderer;
  gboolean is_first;
  gboolean first_point_smooth;
  gboolean last_smooth;
  double last_x, last_y;
};

static void
draw_point (DrawPointsState *state,
	    gboolean         smooth_entry,
	    gboolean         smooth_exit)
{
  FontRenderer *renderer = state->renderer;
  cairo_t *context = renderer->context;
  double radius = renderer->point_radius;

  if (smooth_entry && smooth_exit)
    {
      cairo_arc (context,
		 state->last_x, state->last_y,
		 radius, 0, 2 * M_PI);
    }
  else
    {
      radius /= sqrt(2);
      cairo_rectangle (context,
		       state->last_x - radius, state->last_y - radius,
		       2 * radius, 2 * radius);
    }
  
  cairo_save (context);
  cairo_stroke (context);
  cairo_restore (context);
  cairo_fill (context);
}

static int
draw_points_move_to (FT_Vector *to,
		     void      *user)
{
  DrawPointsState *state = user;
  FontRenderer *renderer = state->renderer;

  if (!state->is_first)
    draw_point (state, state->last_smooth, state->first_point_smooth);

  state->last_x = SCALE (to->x);
  state->last_y = SCALE (- to->y);
  state->is_first = TRUE;
  
  return 0;
}

static int
draw_points_line_to (FT_Vector *to,
		      void      *user)
{
  DrawPointsState *state = user;
  FontRenderer *renderer = state->renderer;
  double to_x, to_y;

  to_x = SCALE (to->x);
  to_y = SCALE (- to->y);

  if (state->is_first)
    state->first_point_smooth = FALSE;
  else
    draw_point (state, state->last_smooth, FALSE);
    
  state->last_x = to_x;
  state->last_y = to_y;
  state->last_smooth = FALSE;
  state->is_first = FALSE;
  
  return 0;
}

static int
draw_points_conic_to (FT_Vector *control,
		      FT_Vector *to,
		      void      *user)
{
  DrawPointsState *state = user;
  FontRenderer *renderer = state->renderer;
  
  double control_x, control_y;
  double to_x, to_y;

  if (state->is_first)
    state->first_point_smooth = TRUE;
  else
    draw_point (state, state->last_smooth, TRUE);

  control_x = SCALE (control->x);
  control_y = SCALE (- control->y);
  to_x = SCALE (to->x);
  to_y = SCALE (- to->y);

  if (renderer->mode == RENDER_MODE_DRAW_CONTROL)
    {
      draw_control (renderer, renderer->context,
		    state->last_x, state->last_y,
		    control_x, control_y, FALSE);
      draw_control (renderer, renderer->context,
		    to_x, to_y,
		    control_x, control_y, TRUE);
    }

  state->last_x = to_x;
  state->last_y = to_y;
  state->last_smooth = TRUE;
  state->is_first = FALSE;
  
  return 0;
}

static int
draw_points_cubic_to (FT_Vector *control1,
		      FT_Vector *control2,
		      FT_Vector *to,
		      void      *user)
{
  DrawPointsState *state = user;
  FontRenderer *renderer = state->renderer;
  
  double control1_x, control1_y;
  double control2_x, control2_y;
  double to_x, to_y;

  if (state->is_first)
    state->first_point_smooth = TRUE;
  else
    draw_point (state, state->last_smooth, TRUE);

  control1_x = SCALE (control1->x);
  control1_y = SCALE (- control1->y);
  control2_x = SCALE (control2->x);
  control2_y = SCALE (- control2->y);
  to_x = SCALE (to->x);
  to_y = SCALE (- to->y);

  if (renderer->mode == RENDER_MODE_DRAW_CONTROL)
    {
      draw_control (renderer, renderer->context,
		    state->last_x, state->last_y,
		    control1_x, control1_y, TRUE);
      draw_control (renderer, renderer->context,
		    to_x, to_y,
		    control2_x, control2_y, TRUE);
    }

  state->last_x = to_x;
  state->last_y = to_y;
  state->last_smooth = TRUE;
  state->is_first = FALSE;
  
  return 0;
}

static const FT_Outline_Funcs draw_points_funcs = {
  draw_points_move_to,
  draw_points_line_to,
  draw_points_conic_to,
  draw_points_cubic_to,
  0,
  0,
};

static void
draw_points (FontRenderer *renderer,
	     cairo_t      *context,
	     FT_Outline   *outline)
{
  DrawPointsState state;
  
  cairo_save (context);
  current_point_to_origin (context);
  cairo_new_path (context);

  state.renderer = renderer;
  state.is_first = TRUE;
  
  FT_Outline_Decompose (outline, &draw_points_funcs, &state);
  if (!state.is_first)
    draw_point (&state, state.last_smooth, state.first_point_smooth);

  cairo_restore (context);
}

static void
draw_glyph (FontRenderer *renderer,
	    cairo_t      *context,
	    PangoFont    *font,
	    PangoGlyph    glyph)
{
  FT_Face face = pango_fc_font_lock_face (PANGO_FC_FONT (font));
  
  FT_Load_Glyph (face, glyph, get_load_flags (font));

  if (face->glyph->format == FT_GLYPH_FORMAT_OUTLINE)
    {
      switch (renderer->mode)
	{
	case RENDER_MODE_STROKE:
	case RENDER_MODE_FILL:
	  draw_outline (renderer, context, &face->glyph->outline);
	  break;
	case RENDER_MODE_DRAW_POINTS:
	case RENDER_MODE_DRAW_CONTROL:
	  draw_points (renderer, context, &face->glyph->outline);
	  break;
	}
    }
		 
  pango_fc_font_unlock_face (PANGO_FC_FONT (font));
}

static void
draw_glyph_string (FontRenderer     *renderer,
		   cairo_t          *context,
		   PangoFont        *font,
		   PangoGlyphString *glyphs)
{
  gint x_off = 0;
  gint i;
  
  cairo_save (context);
  current_point_to_origin (context);

  for (i = 0; i < glyphs->num_glyphs; i++)
    {
      PangoGlyphInfo *gi = &glyphs->glyphs[i];
      
      if (gi->glyph)
	{
	  int x = x_off + gi->geometry.x_offset;
	  int y = gi->geometry.y_offset;

	  move_to_pango (renderer, context, x, y);
	  draw_glyph (renderer, context, font, gi->glyph);
	}
      
      x_off += glyphs->glyphs[i].geometry.width;
    }
  
 cairo_restore (context);
}

static void 
draw_layout_line (FontRenderer    *renderer,
		  cairo_t         *context,
		  PangoLayoutLine *line)
{
  GSList *tmp_list = line->runs;
  gint x_off = 0;

  cairo_save (context);
  current_point_to_origin (context);
  
  while (tmp_list)
    {
      PangoLayoutRun *run = tmp_list->data;
      PangoRectangle logical_rect;

      tmp_list = tmp_list->next;
      
      pango_glyph_string_extents (run->glyphs, run->item->analysis.font,
				  NULL, &logical_rect);

      move_to_pango (renderer, context, x_off, 0);
      draw_glyph_string (renderer, context, run->item->analysis.font, run->glyphs);
      
      x_off += logical_rect.width;
    }
  
  cairo_restore (context);
}

static void 
draw_pango_layout (FontRenderer *renderer,
		   cairo_t      *context,
		   PangoLayout  *layout)
{
  PangoLayoutIter *iter;
	
  cairo_save (context);
  current_point_to_origin (context);

  iter = pango_layout_get_iter (layout);
	
  do
    {
      PangoRectangle   logical_rect;
      PangoLayoutLine *line;
      int              baseline;
      
      line = pango_layout_iter_get_line (iter);
      
      pango_layout_iter_get_line_extents (iter, NULL, &logical_rect);
      baseline = pango_layout_iter_get_baseline (iter);

      move_to_pango (renderer, context, logical_rect.x, baseline);
      draw_layout_line (renderer, context, line);
    }
  while (pango_layout_iter_next_line (iter));

  pango_layout_iter_free (iter);
	
  cairo_restore (context);
}

FontRenderer *
font_renderer_new  (void)
{
  FontRenderer *renderer = g_new0 (FontRenderer, 1);

  renderer->scale = 1.0;
  renderer->point_radius = 3.0;

  return renderer;
}

void
font_renderer_free (FontRenderer *renderer)
{
  g_free (renderer);
}

void
font_renderer_set_scale (FontRenderer *renderer,
			 double        scale)
{
  renderer->scale = scale;
}

void
font_renderer_set_point_radius (FontRenderer *renderer,
				double        point_radius)
{
  renderer->point_radius = point_radius;
}

void
font_renderer_fill_layout (FontRenderer *renderer,
			   cairo_t      *context,
			   PangoLayout  *layout)
{
  renderer->context = context;
  renderer->mode = RENDER_MODE_FILL;

  draw_pango_layout (renderer, context, layout);
  renderer->context = NULL;
}

void
font_renderer_stroke_layout (FontRenderer *renderer,
			     cairo_t      *context,
			     PangoLayout  *layout)
{
  renderer->context = context;
  renderer->mode = RENDER_MODE_STROKE;
  
  draw_pango_layout (renderer, context, layout);
  renderer->context = NULL;
}

void
font_renderer_draw_layout_points   (FontRenderer *renderer,
				    cairo_t      *context,
				    PangoLayout  *layout,
				    gboolean      include_control)
{
  renderer->context = context;
  renderer->mode = include_control ? RENDER_MODE_DRAW_CONTROL : RENDER_MODE_DRAW_POINTS;
  
  draw_pango_layout (renderer, context, layout);
  renderer->context = NULL;
}

void
font_renderer_draw_layout_pixels (FontRenderer *renderer,
				  cairo_t      *context,
				  PangoLayout  *layout)
{
  PangoRectangle ink_rect;
  PangoRectangle ink_pixels;
  FT_Bitmap bitmap;
  int i, j;
  
  cairo_save (context);
  current_point_to_origin (context);
  cairo_new_path (context);

  pango_layout_get_extents (layout, &ink_rect, NULL);
  
  ink_pixels.x = ink_rect.x >> 10;
  ink_pixels.width = ((ink_rect.x + ink_rect.width + 1023) >> 10) - ink_pixels.x;
  
  ink_pixels.y = ink_rect.y >> 10;
  ink_pixels.height = ((ink_rect.y + ink_rect.height + 1023) >> 10) - ink_pixels.y;

  bitmap.width = ink_pixels.width;
  bitmap.pitch = (bitmap.width + 3) & ~3;
  bitmap.rows = ink_pixels.height;
  bitmap.buffer = g_malloc (bitmap.pitch * bitmap.rows);
  bitmap.num_grays = 256;
  bitmap.pixel_mode = FT_PIXEL_MODE_GRAY;
  memset (bitmap.buffer, 0x00, bitmap.pitch * bitmap.rows);

  pango_ft2_render_layout (&bitmap, layout, - ink_pixels.x, - ink_pixels.y);
  
  for (j = 0; j < ink_pixels.height; j++)
    {
      guchar *p = bitmap.buffer + bitmap.pitch * j;
      guchar last_color = 0;	/* Quiet GCC */
      gboolean have_points = FALSE;
      
      for (i = 0; i < ink_pixels.width; i++, p++)
	{
	  if (*p == 0)
	    continue;

	  if (!have_points || *p != last_color)
	    {
	      if (have_points)
		cairo_fill(context);

	      cairo_set_source_rgb (context,
				    (255 - *p) / 255.,
				    (255 - *p) / 255.,
				    (255 - *p) / 255.);
	      last_color = *p;
	      have_points = TRUE;
	    }

	  cairo_rectangle (context,
			   (ink_pixels.x + i) * renderer->scale,
			   (ink_pixels.y + j) * renderer->scale,
			   renderer->scale, renderer->scale);
	}

      if (have_points)
	cairo_fill (context);
    }

  g_free (bitmap.buffer);
  cairo_restore (context);
}
